//     ======================= AssociativeMemory.h ==========================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 03.02.2010
//     **********************************************************************

#if !defined(_AssociativeMemory_H_INCLUDED_)
#define _AssociativeMemory_H_INCLUDED_

#include "cv.h"
#include "cxmisc.h"

#pragma warning (disable:4786) // Identifier was truncated to '255' characters in the debug information

#ifdef AVM_SDK_EXPORTS
#define AVM_SDK_API __declspec(dllexport)
#else
#define AVM_SDK_API __declspec(dllimport)
#endif

#pragma pack(push)
#pragma pack(1)

#define cAM_SimilarityUpTrd   0.54f  // Threshold of similarity on top
#define cAM_SimilarityDnTrd   0.45f  // Threshold of similarity on bottom

// Definition of parameter types
typedef enum {
	ptOptimizeLoop = 0,      // Period of optimization, ms (45000)
	ptTrackingDepth,         // Depth of tracking (40)
	ptTrcDist2Max,           // Threshold of square distance for neighbor objects
	ptExptectDistEr,         // Expected distance error for tracking
	ptRcgMxCellBase,         // Base cell value for recognition threshold calculation (20000)
	ptMaxClusterAge,         // Age of formed cluster (40)
	ptRcgMxExpRate_A,        // | Exponent rate of recognition matrix for learning (0.07)
	ptRcgMxExpRate_B,        // | (0.2)
	ptRcgMxExpRate_C,        // | (0.0015)
	ptSearchScaleStep,       // Scale step of searching  for recognition (0.25)
} CvAM_ParamType;

// Definition of the state constants for recognition function
typedef enum {
	cProcessingIsBegun = 0,
	cUnknownObject,
	cRecognizedObject,
	cGeneralizedObject,
	cTrackedObject,
	cLearnThisObject,
	cProcessingIsFinished,
	cDeletingTrackedObject
} CvAM_State;

// Parameters structure for recognition function
struct CvAM_Parameters {
	CvRect   ObjRect;            // Rectangle where object found
	double   Similarity;         // Similarity of the interest area to an object (0 ... 1)
	long     TotalABases;        // Total number of associative bases
	void**   ppUserTrackingInfo; // Pointer to the tracking info data written by user
};

// Structure of cluster statistics
struct CvClusterStat {
	uint64	       HitTotal;         // Total number of hitting
	double         ActualTick;       // Actual tick number
	double         LastTickDiff;     // Last tick difference after searching
	long           StepsForSearch;   // Steps needed for search 
};

// Definition of coordinate structure
struct CvXYZ32F {
	float x,y,z;
    // Set a coordinate
	void Set(float aX, float aY, float aZ) { x = aX; y = aY; z = aZ; }
};

// Definition of location structure
struct CvLocation {
	CvXYZ32F p; // Location point
	float   hA; // Horizontal survey angle
	float   vA; // Vertical survey angle
};

// Definition of data type CvType160U
#define cType160U_c 20
#define cType160U_s 10
#define cType160U_f 5

union CvType160U {
	char  c[cType160U_c];
	short s[cType160U_s];
	float f[cType160U_f];
	CvLocation l;
};

// Definition of data type CvType192U
#define cType192U_c 24
#define cType192U_s 12
#define cType192U_f 6
#define cType192U_p 2

union CvType192U {
	char  c[cType192U_c];
	short s[cType192U_s];
	float f[cType192U_f];
	CvXYZ32F p[cType192U_p];
};

// Definition of matrix cluster operation function
typedef bool (*TYPE_RcgMCFunc16S)(CvSize  aRcgMxSize, short* apRcgMx,
		                          uint64 aIndex, CvClusterStat* apStat,
								  int aContainBasesTotal, short aLevel, void* args);
typedef TYPE_RcgMCFunc16S CvRcgMCFunc16S;

// Definition of object recognition function for type "long"
typedef void (*TYPE_RcgAMFunc32S)(CvAM_State aState, long* apData, CvAM_Parameters* apParam, void* args);
typedef TYPE_RcgAMFunc32S CvRcgAMFunc32S;

// Definition of associative memory operation function for type "long"
typedef bool (*TYPE_AMFunc32S)(CvSize  aRcgMxSize, short* apRcgMx, long* apData,
							   uint64 aIndex, uint64 aHitCounter,
							   int aContainBasesTotal, short aLevel, void* args);
typedef TYPE_AMFunc32S CvAMFunc32S;

// Definition of object recognition function for type "CvType160U"
typedef void (*TYPE_RcgAMFunc160U)(CvAM_State aState, CvType160U* apData, CvAM_Parameters* apParam, void* args);
typedef TYPE_RcgAMFunc160U CvRcgAMFunc160U;

// Definition of associative memory operation function for type "CvType160U"
typedef bool (*TYPE_AMFunc160U)(CvSize  aRcgMxSize, short* apRcgMx, CvType160U* apData,
								uint64 aIndex, uint64 aHitCounter,
								int aContainBasesTotal, short aLevel, void* args);
typedef TYPE_AMFunc160U CvAMFunc160U;

// Definition of object recognition function for type "CvType192U"
typedef void (*TYPE_RcgAMFunc192U)(CvAM_State aState, CvType192U* apData, CvAM_Parameters* apParam, void* args);
typedef TYPE_RcgAMFunc192U CvRcgAMFunc192U;

// Definition of associative memory operation function for type "CvType192U"
typedef bool (*TYPE_AMFunc192U)(CvSize  aRcgMxSize, short* apRcgMx, CvType192U* apData,
								uint64 aIndex, uint64 aHitCounter,
								int aContainBasesTotal, short aLevel, void* args);
typedef TYPE_AMFunc192U CvAMFunc192U;


// Handle of AVM API
typedef ushort avmHandle;

// Open AVM API
AVM_SDK_API avmHandle __stdcall avmOpen(short aDataSize);

// Close AVM API
AVM_SDK_API void __stdcall avmClose(avmHandle aHandle);

// Reading from associative memory cell (associative base) for type "long"
AVM_SDK_API bool __stdcall avmRead(avmHandle aHandle, CvRect aInterestArea, CvRect* apSearchArea,
	         long** appData, uint64* apIndex, uint64* apHitCounter, double* apSimilarity, bool aTotalSearch);

// Reading from associative memory cell (associative base) for type "CvType160U"
AVM_SDK_API bool __stdcall avmRead(avmHandle aHandle, CvRect aInterestArea, CvRect* apSearchArea,
	         CvType160U** appData, uint64* apIndex, uint64* apHitCounter, double* apSimilarity, bool aTotalSearch);

// Reading from associative memory cell (associative base) for type "CvType192U"
AVM_SDK_API bool __stdcall avmRead(avmHandle aHandle, CvRect aInterestArea, CvRect* apSearchArea,
	         CvType192U** appData, uint64* apIndex, uint64* apHitCounter, double* apSimilarity, bool aTotalSearch);

// Writing to associative memory cell for type "long"
AVM_SDK_API void __stdcall avmWrite(avmHandle aHandle, CvRect aInterestArea, long* apData, bool aDeepLearning);

// Writing to associative memory cell for type "CvType160U"
AVM_SDK_API void __stdcall avmWrite(avmHandle aHandle, CvRect aInterestArea, CvType160U* apData, bool aDeepLearning);

// Writing to associative memory cell for type "CvType192U"
AVM_SDK_API void __stdcall avmWrite(avmHandle aHandle, CvRect aInterestArea, CvType192U* apData, bool aDeepLearning);

// Object recognition
AVM_SDK_API void __stdcall avmObjectRecognition(avmHandle aHandle, CvRcgAMFunc32S RcgAMFunc,  void *args);
AVM_SDK_API void __stdcall avmObjectRecognition(avmHandle aHandle, CvRcgAMFunc160U RcgAMFunc, void *args);
AVM_SDK_API void __stdcall avmObjectRecognition(avmHandle aHandle, CvRcgAMFunc192U RcgAMFunc, void *args);

// Object recognition and tracking
AVM_SDK_API void __stdcall avmObjectTracking(avmHandle aHandle, CvRcgAMFunc32S RcgAMFunc, void *args,
                            bool aTrainingDuringTracking, double aSimilarityUpTrd,double aSimilarityDnTrd);
AVM_SDK_API void __stdcall avmObjectTracking(avmHandle aHandle, CvRcgAMFunc160U RcgAMFunc, void *args,
						    bool aTrainingDuringTracking, double aSimilarityUpTrd,double aSimilarityDnTrd);
AVM_SDK_API void __stdcall avmObjectTracking(avmHandle aHandle, CvRcgAMFunc192U RcgAMFunc, void *args,
						    bool aTrainingDuringTracking, double aSimilarityUpTrd,double aSimilarityDnTrd);

// Perform "AMFunc" for each associative base
AVM_SDK_API bool __stdcall avmDoForEach(avmHandle aHandle, CvAMFunc32S AMFunc,  void *args);
AVM_SDK_API bool __stdcall avmDoForEach(avmHandle aHandle, CvAMFunc160U AMFunc, void *args);
AVM_SDK_API bool __stdcall avmDoForEach(avmHandle aHandle, CvAMFunc192U AMFunc, void *args);

// Creating of associative memory
AVM_SDK_API void __stdcall avmCreate(avmHandle aHandle, CvSize aKeyImgSize,
								     short aLevelMax, int aTreeTotal, bool aClustering);
// Destroying of associative memory
AVM_SDK_API void __stdcall avmDestroy(avmHandle aHandle);

// Set the active associative tree
AVM_SDK_API void __stdcall avmSetActiveTree(avmHandle aHandle, int aTreeIdx);

// Removing all data out from associative tree
AVM_SDK_API void __stdcall avmClearTreeData(avmHandle aHandle);

// Set an image for processing
AVM_SDK_API void __stdcall avmSetImage(avmHandle aHandle, const void* apSrcImg, const void* apObjMaskImg);

// Optimization of associative tree
AVM_SDK_API void __stdcall avmOptimizeAssociativeTree(avmHandle aHandle);

// Get size of packed recognition data
AVM_SDK_API long __stdcall avmGetPackedDataSize(avmHandle aHandle);

// Writing a packed recognition data to memory
AVM_SDK_API void __stdcall avmWritePackedData(avmHandle aHandle, void* apMem);

// Reading a packed recognition data from memory
AVM_SDK_API void __stdcall avmReadPackedData(avmHandle aHandle, void* apMem);

// Saving of recognition data
AVM_SDK_API bool __stdcall avmSave(avmHandle aHandle, char* aFileName, bool aSaveWithoutOptimization);

// Loading of recognition data
AVM_SDK_API bool __stdcall avmLoad(avmHandle aHandle, char* aFileName);

// The estimation of an opportunity for training
AVM_SDK_API double __stdcall avmEstimateOpportunityForTraining(avmHandle aHandle, CvRect aInterestArea);

// Restart time counter of optimization event
AVM_SDK_API void __stdcall avmRestartTimeForOptimization(avmHandle aHandle);

// Perform "RcgMCFunc" for each node of associative bases
AVM_SDK_API bool __stdcall avmDoForEachNode(avmHandle aHandle, CvRcgMCFunc16S RcgMCFunc, void *args);

// Get base level for nodes
AVM_SDK_API short __stdcall avmGetBaseNodeLevel(avmHandle aHandle, short aABaseLevel);

// Get total number of clusters
AVM_SDK_API long __stdcall avmGetClustersTotal(avmHandle aHandle);

// Get total number of node levels
AVM_SDK_API short __stdcall avmGetNodeTotalLevels(avmHandle aHandle);

// Get current index of associative base
AVM_SDK_API uint64 __stdcall avmGetCurIndex(avmHandle aHandle);

// Get total number of associative bases
AVM_SDK_API long __stdcall avmGetTotalABases(avmHandle aHandle);

// Get total number of memory levels
AVM_SDK_API short __stdcall avmGetTotalLevels(avmHandle aHandle);

// Get read-write counter
AVM_SDK_API uint64 __stdcall avmGetWrRdCounter(avmHandle aHandle);

// Get size of key image of associative memory
AVM_SDK_API CvSize __stdcall avmGetKeyImageSize(avmHandle aHandle);

// Get base key size
AVM_SDK_API CvSize __stdcall avmGetBaseKeySize(avmHandle aHandle);

// Set the value of parameter
AVM_SDK_API void __stdcall avmSetParam(avmHandle aHandle, CvAM_ParamType aParam, double aValue);



// Associative video memory consist of associative bases tree and use input
// image as a key for data access.


// Associative memory class
template <class AVM_DataType, class AVM_RcgAMFunc, class AVM_AMFunc>
class CvAssociativeMemory {
	avmHandle hAVM;
public:

	CvAssociativeMemory(int = 0) {
		hAVM = avmOpen(sizeof(AVM_DataType));
	}

	~CvAssociativeMemory() { avmClose(hAVM); }
	
	// Creating of associative memory
	//
	// Notes: The parameter aKeyImgSize is depended on sequence 40, 80, 160, 320, 640, 1280... 2^n*10.
	//        Use of this numbers for sizing gives more accuracy in recognition.
	//        The parameter aLevelMax set a maximal level for associative tree.
	//        If set to zero then a maximal level will be computed as optimal for image key size.
	//        aTreeTotal  - total number of independent associative trees.
	//        aClustering - flag of using cluster tree
	void Create(CvSize aKeyImgSize, short aLevelMax = 0,
		        int aTreeTotal = 1, bool aClustering = true) {
		avmCreate(hAVM, aKeyImgSize,aLevelMax, aTreeTotal, aClustering);
	}
	// Destroying of associative memory
	void Destroy() { avmDestroy(hAVM); }

	// Set the active associative tree
	void SetActiveTree(int aTreeIdx) { avmSetActiveTree(hAVM, aTreeIdx); }
	// Removing all data out from associative tree
	void ClearTreeData() { avmClearTreeData(hAVM); }

	// Set an image for processing
	void SetImage(const void* apSrcImg, const void* apObjMaskImg = NULL) {
		avmSetImage(hAVM, apSrcImg,apObjMaskImg);
	}

	// Reading from associative memory cell (associative base)
	//
	// Notes: aInterestArea - interest area for object searching;
	//        apObjRect     - rectangle where object found;
	//        appData       - pointer to the data of associative cell;
	//        apIndex       - unique index of associative base;
	//        apHitCounter  - number of hitting to associative base;
	//        apSimilarity  - similarity of the interest area to an object (0 ... 1);
	//        aTotalSearch  - flag of the total search until end level of associative memory.
	bool Read(CvRect aInterestArea, CvRect* apObjRect, AVM_DataType** appData,
		      uint64* apIndex = NULL, uint64* apHitCounter = NULL,
			  double* apSimilarity = NULL, bool aTotalSearch = false) {
		return avmRead(hAVM, aInterestArea, apObjRect, appData,
			           apIndex, apHitCounter, apSimilarity, aTotalSearch);
	}

	// Writing to associative memory cell
	void Write(CvRect aInterestArea, AVM_DataType* apData, bool aDeepLearning = true) {
		avmWrite(hAVM, aInterestArea, apData, aDeepLearning);
	}

	// Object recognition
	void ObjectRecognition(AVM_RcgAMFunc RcgAMFunc, void *args = NULL) {
		avmObjectRecognition(hAVM, RcgAMFunc, args);
	}

	// Object recognition and tracking
	void ObjectTracking(AVM_RcgAMFunc RcgAMFunc, void *args = NULL,
		                bool   aTrainingDuringTracking = true,
		                double aSimilarityUpTrd = cAM_SimilarityUpTrd,
						double aSimilarityDnTrd = cAM_SimilarityDnTrd) {
		avmObjectTracking(hAVM, RcgAMFunc, args,
			              aTrainingDuringTracking, aSimilarityUpTrd, aSimilarityDnTrd);
	}

	// Optimization of associative tree
	void OptimizeAssociativeTree() { avmOptimizeAssociativeTree(hAVM); }
	// Restart time counter of optimization event
	void RestartTimeForOptimization() { avmRestartTimeForOptimization(hAVM); }
	
	// Get size of packed recognition data
	long GetPackedDataSize() { return avmGetPackedDataSize(hAVM); }
	// Writing a packed recognition data to memory
	void WritePackedData(void* apMem) { avmWritePackedData(hAVM, apMem); }
	// Reading a packed recognition data from memory
	void ReadPackedData(void* apMem) { avmReadPackedData(hAVM, apMem); }
	
	// Saving of recognition data
	bool Save(char* aFileName, bool aSaveWithoutOptimization = false) {
		return avmSave(hAVM, aFileName, aSaveWithoutOptimization);
	}
	// Loading of recognition data
	bool Load(char* aFileName) { return avmLoad(hAVM, aFileName); }

	// The estimation of an opportunity for training
	double EstimateOpportunityForTraining(CvRect aInterestArea) {
		return avmEstimateOpportunityForTraining(hAVM, aInterestArea);
	}

	// Perform "AMFunc" for each associative base
	bool DoForEach(AVM_AMFunc AMFunc, void *args) { return avmDoForEach(hAVM, AMFunc, args); }
	// Perform "RcgMCFunc" for each node of associative bases
	bool DoForEachNode(CvRcgMCFunc16S RcgMCFunc, void *args) {
		return avmDoForEachNode(hAVM, RcgMCFunc, args);
	}
	// Get base level for nodes
	short GetBaseNodeLevel(short aABaseLevel) { return avmGetBaseNodeLevel(hAVM, aABaseLevel); }
	// Get total number of clusters
	long GetClustersTotal() { return avmGetClustersTotal(hAVM); }
	// Get total number of node levels
	short GetNodeTotalLevels() { return avmGetNodeTotalLevels(hAVM); }
	
	// Get current index of associative base
	uint64 GetCurIndex() { return avmGetCurIndex(hAVM); }
	// Get total number of associative bases
	long   GetTotalABases() { return avmGetTotalABases(hAVM); }
	// Get total number of memory levels
	short  GetTotalLevels() { return avmGetTotalLevels(hAVM); }
	// Get read-write counter
	uint64 GetWrRdCounter() { return avmGetWrRdCounter(hAVM); }
	// Get size of key image of associative memory
	CvSize GetKeyImageSize() { return avmGetKeyImageSize(hAVM); }
	// Get base key size 
	CvSize GetBaseKeySize() { return avmGetBaseKeySize(hAVM); }
	// Set the value of parameter 
	void SetParam(CvAM_ParamType aParam, double aValue) {
		avmSetParam(hAVM, aParam, aValue);
	}
};

// AVM type definition
typedef CvAssociativeMemory<long,       CvRcgAMFunc32S,  CvAMFunc32S>  CvAssociativeMemory32S;
typedef CvAssociativeMemory<CvType160U, CvRcgAMFunc160U, CvAMFunc160U> CvAssociativeMemory160U;
typedef CvAssociativeMemory<CvType192U, CvRcgAMFunc192U, CvAMFunc192U> CvAssociativeMemory192U;

#pragma pack(pop)
#pragma pack()

#endif // !defined(_AssociativeMemory_H_INCLUDED_)
