//     =========================== Navigator.h ==============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 10.01.2010
//     **********************************************************************


#if !defined(_Navigator_H_INCLUDED_)
#define _Navigator_H_INCLUDED_

#ifdef NAVIGATOR_EXPORTS
#define NAVIGATOR_API __declspec(dllexport)
#else
#define NAVIGATOR_API __declspec(dllimport)
#endif

#include "cv.h"
#include "cxmisc.h"

#define cCmdTotal 8
// Array structure of control keys
struct TCmdArray {
	bool Cmd[cCmdTotal];
};

// Constants of robot control
enum {
	cmLEFT = 0,
		cmBACKWARDS,
		cmFORWARD,
		cmRIGHT,
		cmTURRET_RIGHT,
		cmFIRE,
		cmTURRET_LEFT,
		cmPOWER
};

extern NAVIGATOR_API TCmdArray* gpCmdArray; // Robot command buffer
extern NAVIGATOR_API bool gNovaGateMode;    // "Nova gate" mode flag
extern NAVIGATOR_API ushort gCheckpointCounter; // Counter of checkpoint setting
extern NAVIGATOR_API int gWayIdx;               // Route index

#pragma pack(push)
#pragma pack(1)

// Structure of global navigation data
struct TNavigateData {
	int  TurretBalance; // Robot turret balance
	int  MotionSwitch;  // Motion balance switch
};

#define cObjTotalMax 100

// Structure of table recognized object
struct TRcgObj {
	int    ObjTotal;              // Total number of objects
	long   ObjIdx[cObjTotalMax];  // Object index
	CvRect ObjRect[cObjTotalMax]; // Rectangle coordinates of object
	double GateHorizon;           // Gate horizon
	bool   CheckpointFound;       // Flag that set if checkpoint is found
};

#pragma pack(pop)
#pragma pack()

// Definition of log messages function
typedef void (*TYPE_LogMessageFunc)(char* apMsg);
typedef TYPE_LogMessageFunc TLogMessageFunc;

// Setting of checkpoint
typedef void (*TYPE_SetCheckpointFunc)(int, ushort);
typedef TYPE_SetCheckpointFunc TSetCheckpointFunc;

// Robot navigation function
NAVIGATOR_API void Navigate(IplImage* apGray, IplImage* apScreen, TRcgObj* apRcgObj,
							TSetCheckpointFunc SetCheckpoint, TLogMessageFunc LogMessage);
// Starting of navigation mode
NAVIGATOR_API void StartNavigateMode(bool aNovaGateMode = false);
// Ending of navigation mode
NAVIGATOR_API void FinishNavigateMode();

#endif // !defined(_Navigator_H_INCLUDED_)